/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: vdump.c,v 1.16 2009/07/09 02:55:26 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

const char pgoblin_regn[] =
"0123456789"
"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
"abcdefghijklmnopqrstuvwxyz"
"_-"
;

void
pgoblin_Vdump(pgoblin_main *options, pgoblin_prog *pgm) {
    ssize_t p;
    int i;
    u_int32_t t, cmd;
    char *lit;

    ifBLIN_QV4(options->flags) {
        if  (pgm) fprintf(stderr, "+pgoblin_Vdump pgm=%d\n", (int)pgm->curr);
          else fprintf( stderr, "+pgoblin_Vdump pgm=NULL\n");
    }
    printf("+######################################");
    for (p = 0; p < PGOBLIN_STRING_ID_LEN && pgm->id[p]; p++) printf("%c", pgm->id[p]);
    for (p = 0; p < pgm->curr; p++) {
        t = pgm->code[p].cmd;
        cmd = t & PGOBLIN_COMMAND;
        if  (cmd == PGOBLIN_extended) {
            cmd = (pgm->code[p].e.cmd & PGOBLIN_COMMAND) | PGOBLIN_EXTENDED;
            lit = NULL;
        } else {
            lit = pgm->code[p].e.lit;
        }
        if  (cmd >= pgoblin.max && !(~pgoblin.a[cmd].flags & PGOBLIN_ILLEGAL)) {
            printf("\n#ILLEGAL ");
        } else {
            printf("\n#%s ", pgoblin.a[cmd].str);
        }
        t >>= 2;
        for (i = 0; i < pgoblin.binparam; i++)
            printf("%c", pgoblin_regn[(t >>= PGOBLIN_BPARG) & PGOBLIN_ARGMASK]);
        printf(" ([%d]=%03X)", (int)p, cmd);
        if  (lit) printf("\n%s", pgm->code[p].e.lit);
    }
    printf("\n-######################################\n");
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_Vdump\n");
};

void
pgoblin_VVdump() {
    int i;

    printf( "SYMREGS=%d BINREGS=%d PGOBLIN_max=%03X\n"
          , (int)pgoblin.symparam, (int)pgoblin.binparam, pgoblin.max
          );
    for (i = 0; i < pgoblin.binparam; i++)
        printf("ARG%dS=%d%s", i + 1, pgoblin.binparams[i], (i + 1 < pgoblin.binparam) ? " " : "\n");
    printf("sym val bin mask name\n");
    for (i = 0; i < pgoblin.symparam; i++)
        printf( " %02X  %02X %s%2d %4d %s\n"
              , pgoblin.p[i].sym
              , pgoblin.p[i].val
              , (pgoblin.p[i].bin < 0) ? "~" : " "
              , (pgoblin.p[i].bin < 0) ? ~pgoblin.p[i].bin : pgoblin.p[i].bin
              , pgoblin.p[i].mask
              , pgoblin.p[i].name
              );
    printf("      flags     car   str\n");
    for (i = 0; i < pgoblin.max; i++) printf( "%03X %08X %" BLIN_X "%s%s%s\n"
                                            , i
                                            , pgoblin.a[i].flags
                                            , BLIN_I(pgoblin.a[i].car)
                                            , (pgoblin.a[i].str) ? " =" : ""
                                            , (pgoblin.a[i].str) ? pgoblin.a[i].str : ""
                                            , (pgoblin.a[i].str) ? "~" : ""
                                            );
};

static int
cmp(const void *a, const void *b) {
    return(strcmp(pgoblin.a[*(int*)a].str, pgoblin.a[*(int*)b].str));
}

void
pgoblin_htman(u_int32_t flags, char *Tflag) {
    switch(Tflag[0]) {
    case '1': {
            u_int32_t tag[PGOBLIN_SYMREGS];
            char *pt[PGOBLIN_SYMREGS];
            u_int8_t p[PGOBLIN_BINREGS];
            int ex = EX_OK, i, j, k, m;
            int *sqn;
            char cf[8];

        ;   if  (!(sqn = malloc(sizeof(int) * pgoblin.max))) {
        ;       errno = ENOMEM;
        ;       ex = EX_UNAVAILABLE;
        ;       ifBLIN_QV1(flags) warn("pgoblin_htman:malloc");
        ;       goto out;
        ;   }
        ;   for (i = 1, j = 0; i < pgoblin.max; i++) {
        ;       if  (pgoblin.a[i].str) sqn[j++] = i;
        ;   }
        ;   if  (heapsort(sqn, j, sizeof(int), cmp) < 0) {
        ;       ex = EX_SOFTWARE;
        ;       ifBLIN_QV1(flags) warn("radixsort");
        ;       goto out;
        ;   }
        ;   for (i = 0; i < j; i++) {
                u_int32_t f;
                u_int8_t h;
                char c;

        ;       f = pgoblin.a[sqn[i]].flags;
        ;       if  ((f & PGOBLIN_ILLEGAL) == PGOBLIN_ILLEGAL) {
        ;           tag[0] = 8;
        ;           f &= ~PGOBLIN_ILLEGAL;
        ;       } else {
        ;           for (k = 0; k < PGOBLIN_BINREGS; k++) p[k] = 0;
        ;           for (k = 0; k < PGOBLIN_SYMREGS; k++) {
        ;               tag[k] = 0;
        ;               if  (f & pgoblin.p[k].sym) {
                            int r;

        ;                   switch ((h = pgoblin.a[sqn[i]].ch[r = PREG(sqn[i], k)])) {
        ;                   case 0                   : tag[k] |= 4; pt[k] = "";      break;
        ;                   case PGOBLIN_PATHPARM                 : pt[k] = "Pf";    break;
        ;                   case PGOBLIN_TEXTPARM                 : pt[k] = "Pt";    break;
        ;                   case PGOBLIN_MIFEDES                  : pt[k] = "Pm";    break;
        ;                   case PGOBLIN_PQRESULT                 : pt[k] = "Pq";    break;
        ;                   case PGOBLIN_STYLE                    : pt[k] = "Ps";    break;
        ;                   case PGOBLIN_JOB                      : pt[k] = "Pj";    break;
        ;                   case PGOBLIN_CONNECT                  : pt[k] = "Pc";    break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_PATHPARM: pt[k] = "Po Pf"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_TEXTPARM: pt[k] = "Po Pt"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_MIFEDES : pt[k] = "Po Pm"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_PQRESULT: pt[k] = "Po Pq"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_STYLE   : pt[k] = "Po Ps"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_JOB     : pt[k] = "Po Pj"; break;
        ;                   case PGOBLIN_OUTSET | PGOBLIN_CONNECT : pt[k] = "Po Pc"; break;
        ;                   default: tag[k] |= 1; asprintf(&(pt[k]), "%d", h);
        ;                   }
//printf("%d s%d b%d t%c %02X\n", tag[k], k, r, (!p[r]) ? '~' : ('0' + ~p[r]), h);
        ;                   if  (h & ~(PGOBLIN_OUTSET | pgoblin.p[k].val)) tag[k] |= 1;
        ;                   if  (!(h & pgoblin.p[k].val) != !*(pt[k])) tag[k] |= 1;
        ;                   f &= ~pgoblin.p[k].sym;
        ;                   if  (p[r]) {
        ;                       tag[~p[r]] |= 2;
        ;                       tag[k] |= 2;
        ;                   }
        ;                   p[r] = ~k;
        ;               } else {
        ;                   pt[k] = "";
        ;               }
        ;               if  (pgoblin.p[k].bin < 0) f &= ~(pgoblin.p[k].mask << ~pgoblin.p[k].bin);
        ;       }   }
        ;       k = 0;
        ;       c = ' ';
        ;       if  (f & PGOBLIN_IGNORE) {
        ;           cf[k++] = 'i';
        ;           for (m = 0; m < PGOBLIN_BINREGS; m++) if (p[m]) c = '!';
        ;           for (m = 0; m < PGOBLIN_SYMREGS; m++) if (*(pt[m])) c = '!';
        ;           f &= ~PGOBLIN_IGNORE;
        ;       }
        ;       if  (f & PGOBLIN_NOLITER) {
        ;           cf[k++] = 'l';
        ;           if (p[1]) c = '!';
        ;           if (tag[1] || *(pt[1])) c = '!';
        ;           f &= ~PGOBLIN_NOLITER;
        ;       }
        ;       if  (f & PGOBLIN_LEXLVUP) cf[k++] = 'u';
        ;       if  (f & PGOBLIN_LEXLDOWN) cf[k++] = 'd';
        ;       if  (f & PGOBLIN_FORKED) cf[k++] = 'f';
        ;       f &= ~(PGOBLIN_LEXLVUP | PGOBLIN_LEXLDOWN | PGOBLIN_FORKED);
        ;       if  (f) c = '!';
        ;       for (m = 0; m < PGOBLIN_SYMREGS; m++) if (tag[m] & ~8) c = '!';
        ;       if  (c == '!') cf[k++] = '!';
        ;       cf[k++] = '\0';
        ;       ifBLIN_QV2(flags) printf( "%08X%c\t", f, c);
        ;       if  (tag[0] & 8) printf( "%s\t%s\t%d\t\t0\t\t0\t\t0\t\t0\t\t0\t\n"
                                       , cf, pgoblin.a[sqn[i]].str, tag[0]
                                       );
                  else printf( "%s\t%s\t%d\t%s\t%d\t%s\t%d\t%s\t%d\t%s\t%d\t%s\t%d\t%s\n"
                                 , cf, pgoblin.a[sqn[i]].str
                             , tag[0], pt[0]
                             , tag[1], pt[1]
                             , tag[2], pt[2]
                             , tag[3], pt[3]
                             , tag[4], pt[4]
                             , tag[5], pt[5]
                             );
        }   }
        break;
    default: {
            char *p;

            for (p = Tflag; *p; p++) {
                if  (p[1] == ':') printf("%c\t1\n", *(p++));
                  else printf("%c\t0\n", *p);
    }   }   }
out:;
}
